//Pipelined AXI4 Master (Burst capable)
//Implements full AXI4 interface, as well as AXI-Stream-like
//command and data interfaces on separate clock domains

module axi4_master
#(
    parameter ID_WIDTH   = 2,
    parameter ADDR_WIDTH = 32,
    parameter DATA_WIDTH = 32,
    parameter CMD_SIZE = 10
)
(
//system signals
input axi4_aclk,
input axi4_aresetn,//active low, asynchronous assertion and synchronous deassertion

//Write channels
//write address
input                       axi4_awready,
output reg                  axi4_awvalid,
output reg [ADDR_WIDTH-1:0] axi4_awaddr,
output reg [2:0]            axi4_awprot,
output reg [7:0]            axi4_awlen,
output reg [2:0]            axi4_awsize,
output reg [1:0]            axi4_awburst,
output reg                  axi4_awlock,
output reg [3:0]            axi4_awcache,
output reg [ID_WIDTH-1:0]   axi4_awid,
//write data
input                       axi4_wready,
output                      axi4_wvalid,
output [DATA_WIDTH-1:0]     axi4_wdata,
output [(DATA_WIDTH/8)-1:0] axi4_wstrb,
output                      axi4_wlast,
//write burst response
output                      axi4_bready,
input                       axi4_bvalid,
input [1:0]                 axi4_bresp,
input [ID_WIDTH-1:0]        axi4_bid,

//Read channels
//read address
input                       axi4_arready,
output reg                  axi4_arvalid,
output reg [ADDR_WIDTH-1:0] axi4_araddr,
output reg [2:0]            axi4_arprot,
output reg [7:0]            axi4_arlen,
output reg [2:0]            axi4_arsize,
output reg [1:0]            axi4_arburst,
output reg                  axi4_arlock,
output reg [3:0]            axi4_arcache,
output reg [ID_WIDTH-1:0]   axi4_arid,
//read data
output                      axi4_rready,
input                       axi4_rvalid,
input [1:0]                 axi4_rresp,
input [DATA_WIDTH-1:0]      axi4_rdata,
input                       axi4_rlast,
input [ID_WIDTH-1:0]        axi4_rid,

//AXI-Stream compatible command input interface
input                       cmd_s_axis_aclk,
output                      cmd_s_axis_tready,
input                       cmd_s_axis_tvalid,
input [CMD_SIZE-1:0]        cmd_s_axis_tdata,
input                       cmd_s_axis_tlast,
input                       cmd_s_axis_tuser,

//AXI-Stream compatible command response output interface
input                       cmd_m_axis_aclk,
input                       cmd_m_axis_tready,
output                      cmd_m_axis_tvalid,
output [1:0]                cmd_m_axis_tdata,
output                      cmd_m_axis_tlast,
output                      cmd_m_axis_tuser,

//AXI-Stream compatible data input interface
input                       data_s_axis_aclk,
output                      data_s_axis_tready,
input                       data_s_axis_tvalid,
input [DATA_WIDTH-1:0]      data_s_axis_tdata,
input                       data_s_axis_tlast,
input [(DATA_WIDTH/8)-1:0]  data_s_axis_tuser,

//AXI-Stream compatible data output interface
input                       data_m_axis_aclk,
input                       data_m_axis_tready,
output                      data_m_axis_tvalid,
output [DATA_WIDTH-1:0]     data_m_axis_tdata,
output                      data_m_axis_tlast,
output [ID_WIDTH+2-1:0]     data_m_axis_tuser

);
/*
localparam CMD_SIZE = ADDR_WIDTH+ID_WIDTH+3+8+3+2+1+4+1;
localparam  RW_BIT_OFFSET = 0;
            

//read/write command issue logic 
wire write_request, read_request;
wire cmd_valid;
wire [CMD_SIZE-1:0] cmd_data;

assign write_request = cmd_valid & cmd_data[RW_BIT_OFFSET] & ~axi4_awvalid;
assign read_request = cmd_valid & ~cmd_data[RW_BIT_OFFSET] & ~axi4_awvalid;

always @(posedge axi4_aclk or negedge axi4_aresetn)
    if(~axi4_aresetn) begin
        axi4_awvalid <= 0;
        axi4_arvalid <= 0;
    end else begin
        if(axi4_awvalid & axi4_awready)
            axi4_awvalid <= 0;
        else if(write_request)
            axi4_awvalid <= 1;
        if(axi4_arvalid & axi4_arready)
            axi4_arvalid <= 0;
        else if(read_request)
            axi4_arvalid <= 1;
    end

always @(posedge axi4_aclk) begin
    if(write_request) begin
        axi4_awaddr <= cmd_data[ADDR_BIT_OFFSET+ADDR_WIDTH-1:ADDR_BIT_OFFSET];
        axi4_awprot <= cmd_data[PROT_BIT_OFFSET+3-1:PROT_BIT_OFFSET];
        axi4_awlen <= cmd_data[LEN_BIT_OFFSET+8-1:LEN_BIT_OFFSET];
        axi4_awsize <= cmd_data[SIZE_BIT_OFFSET+3-1:SIZE_BIT_OFFSET];
        axi4_awburst <= cmd_data[BURST_BIT_OFFSET+2-1:BURST_BIT_OFFSET];
        axi4_awlock <= cmd_data[LOCK_BIT_OFFSET];
        axi4_awcache <= cmd_data[CACHE_BIT_OFFSET+4-1:CACHE_BIT_OFFSET];
        axi4_awid <= cmd_data[ID_BIT_OFFSET+ID_WIDTH-1:ID_BIT_OFFSET];
    end
    if(read_request) begin
        axi4_araddr <= cmd_data[ADDR_BIT_OFFSET+ADDR_WIDTH-1:ADDR_BIT_OFFSET];
        axi4_arprot <= cmd_data[PROT_BIT_OFFSET+3-1:PROT_BIT_OFFSET];
        axi4_arlen <= cmd_data[LEN_BIT_OFFSET+8-1:LEN_BIT_OFFSET];
        axi4_arsize <= cmd_data[SIZE_BIT_OFFSET+3-1:SIZE_BIT_OFFSET];
        axi4_arburst <= cmd_data[BURST_BIT_OFFSET+2-1:BURST_BIT_OFFSET];
        axi4_arlock <= cmd_data[LOCK_BIT_OFFSET];
        axi4_arcache <= cmd_data[CACHE_BIT_OFFSET+4-1:CACHE_BIT_OFFSET];
        axi4_arid <= cmd_data[ID_BIT_OFFSET+ID_WIDTH-1:ID_BIT_OFFSET];
    end
end

//command issue FIFO
axis_async_fifo #(
    .TDATA_SIZE(CMD_SIZE),
    .DEPTH_LOG(4),
    .TUSER_SIZE(1'b1)
) command_issue_fifo (
    .aresetn(axi4_aresetn),

    .s_axis_aclk(cmd_s_axis_aclk),
    .s_axis_tvalid(cmd_s_axis_tvalid),
    .s_axis_tdata(cmd_s_axis_tdata),
    .s_axis_tlast(cmd_s_axis_tlast),
    .s_axis_tuser(cmd_s_axis_tuser),
    .s_axis_tready(cmd_s_axis_tready),

    .m_axis_aclk(axi4_aclk),
    .m_axis_tvalid(cmd_valid),
    .m_axis_tdata(cmd_data),
    .m_axis_tlast(),
    .m_axis_tuser(),
    .m_axis_tready(write_request | read_request)
);

//command response FIFO
axis_async_fifo #(
    .TDATA_SIZE(2),
    .DEPTH_LOG(4),
    .TUSER_SIZE(1)
) command_response_fifo (
    .aresetn(axi4_aresetn),

    .s_axis_aclk(axi4_aclk),
    .s_axis_tvalid(axi4_bvalid),
    .s_axis_tdata(axi4_bresp),
    .s_axis_tlast(1'b1),
    .s_axis_tuser(1'b0),
    .s_axis_tready(axi4_bready),

    .m_axis_aclk(cmd_m_axis_aclk),
    .m_axis_tvalid(cmd_m_axis_tvalid),
    .m_axis_tdata(cmd_m_axis_tdata),
    .m_axis_tlast(cmd_m_axis_tlast),
    .m_axis_tuser(cmd_m_axis_tuser),
    .m_axis_tready(cmd_m_axis_tready)
);

//Write data FIFO
axis_async_fifo #(
    .TDATA_SIZE(DATA_WIDTH),
    .DEPTH_LOG(6),
    .TUSER_SIZE(DATA_WIDTH/8)
) write_data_fifo (
    .aresetn(axi4_aresetn),

    .s_axis_aclk(data_m_axis_aclk),
    .s_axis_tvalid(data_s_axis_tvalid),
    .s_axis_tdata(data_s_axis_tdata),
    .s_axis_tlast(data_s_axis_tlast),
    .s_axis_tuser(data_s_axis_tuser),
    .s_axis_tready(data_s_axis_tready),

    .m_axis_aclk(axi4_aresetn),
    .m_axis_tvalid(axi4_wvalid),
    .m_axis_tdata(axi4_wdata),
    .m_axis_tlast(axi4_wlast),
    .m_axis_tuser(axi4_wstrb),
    .m_axis_tready(axi4_wready)
);

//Read data FIFO
axis_async_fifo #(
    .TDATA_SIZE(DATA_WIDTH),
    .DEPTH_LOG(6),
    .TUSER_SIZE(ID_WIDTH+2)
) read_data_fifo (
    .aresetn(axi4_aresetn),

    .s_axis_aclk(axi4_aresetn),
    .s_axis_tvalid(axi4_rvalid),
    .s_axis_tdata(axi4_rdata),
    .s_axis_tlast(axi4_rlast),
    .s_axis_tuser({axi4_rid,axi4_rresp}),
    .s_axis_tready(axi4_rready),

    .m_axis_aclk(data_m_axis_aclk),
    .m_axis_tvalid(data_m_axis_tvalid),
    .m_axis_tdata(data_m_axis_tdata),
    .m_axis_tlast(data_m_axis_tlast),
    .m_axis_tuser(data_m_axis_tuser),
    .m_axis_tready(data_m_axis_tready)
);
*/
endmodule
