module axi4lite_slave_tb;

parameter ADDR_WIDTH = 32;
parameter DATA_WIDTH = 32;
parameter ACLK_PERIOD = 5.55;
parameter IPCLK_PERIOD = 7.77;

//system signals
reg axi4l_aclk;
reg axi4l_aresetn;

//Write channels
//write address
wire                      axi4l_awready;
wire                      axi4l_awvalid;
wire [ADDR_WIDTH-1:0]     axi4l_awaddr;
wire [2:0]                axi4l_awprot;
//write data
wire                      axi4l_wready;
wire                      axi4l_wvalid;
wire [DATA_WIDTH-1:0]     axi4l_wdata;
wire [(DATA_WIDTH/8)-1:0] axi4l_wstrb;
//burst response
wire                      axi4l_bready;
wire                      axi4l_bvalid;
wire [1:0]                axi4l_bresp;//NOTE: 00 = OKAY, 10 = SLVERR (write error)

//Read channels
//read address
wire                      axi4l_arready;
wire                      axi4l_arvalid;
wire [ADDR_WIDTH-1:0]     axi4l_araddr;
wire [2:0]                axi4l_arprot;
//read data
wire                      axi4l_rready;
wire                      axi4l_rvalid;
wire [1:0]                axi4l_rresp;//NOTE: 00 = OKAY, 10 = SLVERR (read error)
wire [DATA_WIDTH-1:0]     axi4l_rdata;

//IP-side interface
reg                       ip_clk;
wire                      ip_ren;
wire                      ip_wen;
wire [ADDR_WIDTH-1:0]     ip_addr;
wire [(DATA_WIDTH/8)-1:0] ip_wstrb;
wire [DATA_WIDTH-1:0]     ip_wdata;
reg                       ip_wack;
reg                       ip_rack;
reg [DATA_WIDTH-1:0]      ip_rdata;
reg                       ip_error;

axi4lite_slave #(
    .ADDR_WIDTH(ADDR_WIDTH),
    .DATA_WIDTH(DATA_WIDTH)
) dut (
    //system signals
    .axi4l_aclk,
    .axi4l_aresetn,
    //Write channels
    //write address
    .axi4l_awready,
    .axi4l_awvalid,
    .axi4l_awaddr,
    .axi4l_awprot,
    //write data
    .axi4l_wready,
    .axi4l_wvalid,
    .axi4l_wdata,
    .axi4l_wstrb,
    //burst response
    .axi4l_bready,
    .axi4l_bvalid,
    .axi4l_bresp,
    //Read channels
    //read address
    .axi4l_arready,
    .axi4l_arvalid,
    .axi4l_araddr,
    .axi4l_arprot,
    //read data
    .axi4l_rready,
    .axi4l_rvalid,
    .axi4l_rresp,
    .axi4l_rdata,
    //IP-side interface
    .ip_clk,
    .ip_ren,
    .ip_wen,
    .ip_addr,
    .ip_wstrb,
    .ip_wdata,
    .ip_wack,
    .ip_rack,
    .ip_rdata,
    .ip_error
);

axi4lite_master_bfm #(
    .COMMAND_FILE("cmd.txt"),
    .ADDR_WIDTH(ADDR_WIDTH),
    .DATA_WIDTH(DATA_WIDTH)
) axi4l_master_bfm (
    //system signals
    .axi4l_aclk,
    .axi4l_aresetn,
    //Write channels
    //write address
    .axi4l_awready,
    .axi4l_awvalid,
    .axi4l_awaddr,
    .axi4l_awprot,
    //write data
    .axi4l_wready,
    .axi4l_wvalid,
    .axi4l_wdata,
    .axi4l_wstrb,
    //burst response
    .axi4l_bready,
    .axi4l_bvalid,
    .axi4l_bresp,
    //Read channels
    //read address
    .axi4l_arready,
    .axi4l_arvalid,
    .axi4l_araddr,
    .axi4l_arprot,
    //read data
    .axi4l_rready,
    .axi4l_rvalid,
    .axi4l_rresp,
    .axi4l_rdata
);

//generate clocks of desired frequency
initial begin
    axi4l_aclk = 0;
    forever #(ACLK_PERIOD/2) axi4l_aclk = ~axi4l_aclk;
end
initial begin
    ip_clk = 0;
    forever #(IPCLK_PERIOD/2) ip_clk = ~ip_clk;
end

//generate reset
initial begin
    axi4l_aresetn = 1;
    #1000
    axi4l_aresetn = 0;
    #1000
    axi4l_aresetn = 1;
end

initial begin
    ip_wack = 0;
    ip_rack = 0;
end

//use small memory to hold written data
//user should be mindful of aliasing
reg [DATA_WIDTH-1:0] mem[1023:0];

//process to respond to writes with ack and random error (random latency 1 up to 4 clocks)
always @(posedge ip_wen) begin
    mem[ip_addr/(DATA_WIDTH/8)] = ip_wdata;
    repeat(1 + ($random % 4)) @(posedge ip_clk);
    ip_wack = 1;
    ip_error = $random;
    @(posedge ip_clk);
    ip_wack = 0;
    ip_error = 0;
end

reg [ADDR_WIDTH-1:0] rdaddr;
//respond to reads with random data and random error (random latency 1 up to 4 clocks)
always @(posedge ip_ren) begin
    rdaddr = ip_addr;
    repeat(1 + ($random % 4)) @(posedge ip_clk);
    ip_rack = 1;
    ip_error = $random;
    ip_rdata = mem[rdaddr/(DATA_WIDTH/8)];
    @(posedge ip_clk);
    ip_rack = 0;
    ip_error = 0;
    ip_rdata = 0;
end



endmodule
