//Bus-functional model for an AXI4-Lite master
//Reads up to 100 commands from a file, executes them over AXI4 interface
//available command mnemonics are:
//WAIT <number of axi4_aclk periods in hex>
//READ <address in hex> <expected data in hex>
//WRITE <address in hex> <data in hex>

module axi4lite_master_bfm
#(
    parameter COMMAND_FILE = "",
    parameter ADDR_WIDTH = 32,
    parameter DATA_WIDTH = 32//AXI4 spec requires this to be strictly 32 or 64
)
(
    //system signals
    input axi4l_aclk,
    input axi4l_aresetn,//active low, asynchronous assertion and synchronous deassertion

    //Write channels
    //write address
    input                           axi4l_awready,
    output reg                      axi4l_awvalid,
    output reg [ADDR_WIDTH-1:0]     axi4l_awaddr,
    output reg [2:0]                axi4l_awprot,
    //write data
    input                           axi4l_wready,
    output reg                      axi4l_wvalid,
    output reg [DATA_WIDTH-1:0]     axi4l_wdata,
    output reg [(DATA_WIDTH/8)-1:0] axi4l_wstrb,
    //burst response
    output reg                      axi4l_bready,
    input                           axi4l_bvalid,
    input [1:0]                     axi4l_bresp,//NOTE: 00 = OKAY, 10 = SLVERR (write error)

    //Read channels
    //read address
    input                           axi4l_arready,
    output reg                      axi4l_arvalid,
    output reg [ADDR_WIDTH-1:0]     axi4l_araddr,
    output reg [2:0]                axi4l_arprot,
    //read data
    output reg                      axi4l_rready,
    input                           axi4l_rvalid,
    input [1:0]                     axi4l_rresp,//NOTE: 00 = OKAY, 10 = SLVERR (read error)
    input [DATA_WIDTH-1:0]          axi4l_rdata,

    output reg                      done
);

integer cmd_file = 0;
reg [100*8-1:0] cmd;
integer arg1, arg2;
integer read_data;

initial begin
    done = 0;
    @(negedge axi4l_aresetn);
    axi4l_awvalid = 0;
    axi4l_wvalid = 0;
    axi4l_arvalid = 0;
    axi4l_bready = 0;
    axi4l_rready = 0;
    @(posedge axi4l_aresetn);
    cmd_file = $fopen(COMMAND_FILE,"r");
    if(cmd_file == 0) begin
        $display("AXI4BFM: Unable to open command file");
        $finish;
    end else
        $display("AXI4LBFM: Opened command file %s",COMMAND_FILE);
    while(!$feof(cmd_file)) begin
        $fscanf(cmd_file,"%s %h %h\n",cmd,arg1,arg2);
        case(cmd)
            "WAIT": begin
                $display("Waiting %d clock cycles",arg1);
                repeat(arg1) @(posedge axi4l_aclk);
            end
            "READ": begin
                $display("Reading address %h, expecting %h",arg1,arg2);
                issue_axi4l_read(arg1,read_data);
                if(arg2 !== read_data) begin
                    $display("Unexpected data read from %h",arg1);
                    $stop;
                end
            end
            "WRITE": begin
                $display("Writing address %h with value %h",arg1,arg2);
                issue_axi4l_write(arg1,arg2,{(DATA_WIDTH/8){1'b1}});
            end
            default:
                $display("Unrecognized command, skipping");
        endcase
    end
    $fclose(cmd_file);
    done = 1;
end



task issue_axi4l_write;
    input [ADDR_WIDTH-1:0] address;
    input [DATA_WIDTH-1:0] data;
    input [DATA_WIDTH/8-1:0] strobe;
    begin
        fork
            begin
                @(negedge axi4l_aclk);
                axi4l_awvalid = 1;
                axi4l_awaddr = address;
                axi4l_awprot = 0;
                if(axi4l_awready)
                    @(posedge axi4l_aclk);
                else while(~axi4l_awready)
                    @(posedge axi4l_aclk);
                axi4l_awvalid = 0;
            end
            begin
                @(negedge axi4l_aclk);
                axi4l_wvalid = 1;
                axi4l_wdata = data;
                axi4l_wstrb = strobe;
                while(~axi4l_wready)
                    @(posedge axi4l_aclk);
                axi4l_wvalid = 0;
            end
            begin
                @(posedge axi4l_bvalid);
                @(negedge axi4l_aclk);
                axi4l_bready = 1;
                @(negedge axi4l_aclk);
                axi4l_bready = 0;
            end
        join
    end
endtask

task issue_axi4l_read;
    input [ADDR_WIDTH-1:0] address;
    output [DATA_WIDTH-1:0] data;
    begin
        fork
            begin
                @(negedge axi4l_aclk);
                axi4l_arvalid = 1;
                axi4l_araddr = address;
                axi4l_arprot = 0;
                if(axi4l_arready)
                    @(posedge axi4l_aclk);
                else while(~axi4l_arready)
                    @(posedge axi4l_aclk);
                while(~axi4l_arready)
                    @(posedge axi4l_aclk);
                axi4l_arvalid = 0;
            end
            begin
                @(posedge axi4l_rvalid);
                @(negedge axi4l_aclk);
                data = axi4l_rdata;
                axi4l_rready = 1;
                @(negedge axi4l_aclk);
                axi4l_rready = 0;
            end
        join
    end
endtask

endmodule
